/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.base.ui.editpart;

import java.util.List;

import org.conqat.lib.commons.string.StringUtils;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.XYLayout;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.Request;
import org.eclipse.gef.RequestConstants;
import org.eclipse.swt.graphics.Color;
import org.fortiss.tooling.base.layout.constants.RecommendedLayoutConstants;
import org.fortiss.tooling.base.layout.util.LayoutDataUtils;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;
import org.fortiss.tooling.kernel.services.IModelElementService;

/**
 * The base class for graphical edit parts which encapsulate nodes (i.e. the
 * main modeling elements). All of them have a name label and an activation
 * mode.
 * 
 * @param <T>
 *            the type of the model element wrapped.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash: D6B33EFB704B0CA151E1C945B5BC48C6
 */
public abstract class NodeEditPartBase<T extends ILayoutedModelElement & INamedCommentedElement>
		extends PositionedEditPartBase<T> {

	/** The label used for the name. */
	protected final Label nameLabel = new Label();

	/** Whether this is active. Also see {@link #setActive(boolean)}. */
	protected boolean active = false;

	/** Constructor. */
	protected NodeEditPartBase(T modelObject) {
		super(modelObject);

		nameLabel.setForegroundColor(getForegroundColor());
		nameLabel.setFont(RecommendedLayoutConstants.TITLE_FONT);
		nameLabel.setTextAlignment(PositionConstants.LEFT
				| PositionConstants.TOP);
		nameLabel.setTextPlacement(PositionConstants.EAST);
		nameLabel.setIcon(IModelElementService.INSTANCE.getModelElementHandler(
				modelElement).getIcon());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected final IFigure createFigure() {
		final IFigure figure = createBaseFigure();
		figure.setBackgroundColor(getBackgroundColor());

		figure.setLayoutManager(new XYLayout());
		figure.add(nameLabel);
		// TODO (FH): implement marker support
		// figure.add(markerFigure);

		return figure;
	}

	/** Creates the outer base figure. */
	protected abstract IFigure createBaseFigure();

	/**
	 * Returns the foreground color used.
	 */
	protected abstract Color getForegroundColor();

	/**
	 * Returns the background color used.
	 */
	protected abstract Color getBackgroundColor();

	/**
	 * Returns the background color for active nodes.
	 */
	protected Color getActiveColor() {
		return getBackgroundColor();
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Returns the connectors as children.
	 */
	@Override
	protected List<?> getModelChildren() {
		return modelElementHandler.getConnectors(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	protected Rectangle determinePosition() {
		return LayoutDataUtils.getNodeBounds(modelElement);
	}

	/**
	 * {@inheritDoc}
	 * <p>
	 * Implements the component role for deletion.
	 */
	@Override
	protected void createEditPolicies() {
		// TODO (FH): add remove policy
		// installEditPolicy(EditPolicy.COMPONENT_ROLE, new RemoveEditPolicy());
	}

	/** {@inheritDoc} */
	@Override
	public void performRequest(Request req) {
		if (req.getType() == RequestConstants.REQ_OPEN) {

			// TODO (FH): add double-click open editor support
			// ModelContext context = ModelContextManager.getInstance()
			// .getContext(modelElement);
			// CCTSEditorInput editorInput = new CCTSEditorInput(modelElement,
			// context);
			// try {
			// PlatformUI.getWorkbench().getActiveWorkbenchWindow()
			// .getActivePage()
			// .openEditor(editorInput, CCTSBindingEditor.ID);
			// } catch (PartInitException e) {
			// LoggingUtils
			// .error(CCTSModelBaseActivator.getDefault(),
			// "Could not open editor with ID "
			// + CCTSBindingEditor.ID, e);
			// }
		}
		super.performRequest(req);
	}

	/** {@inheritDoc} */
	@Override
	protected void refreshVisuals() {
		super.refreshVisuals();

		nameLabel.setText(modelElement.getName());

		IFigure figure = getFigure();

		String tooltipText = StringUtils.wrapLongLines(modelElement.getName()
				+ ": " + modelElement.getComment(), 100);
		figure.setToolTip(new Label(tooltipText));

		figure.setBackgroundColor(active ? getActiveColor()
				: getBackgroundColor());

		Rectangle rect = LayoutDataUtils.getNodeBounds(modelElement);
		int inset = 2 * RecommendedLayoutConstants.RECTANGULAR_SHAPE_INSETS;
		rect = new Rectangle(inset, inset, rect.width - 2 * inset, rect.height
				- 2 * inset);
		figure.setConstraint(nameLabel, rect);
	}

	/**
	 * This method can be used to mark this editpart as active, which can be
	 * used e.g. for displaying the active state in the simulation.
	 */
	public void setActive(boolean active) {
		if (this.active != active) {
			this.active = active;
			refreshVisuals();
		}
	}
}
