/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.base.ui.layout.util;

import java.awt.Polygon;

import org.eclipse.draw2d.PositionConstants;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.fortiss.tooling.base.layout.constants.LayoutKeyConstants;
import org.fortiss.tooling.base.layout.constants.RecommendedLayoutConstants;
import org.fortiss.tooling.base.layout.util.LayoutDataUtils;
import org.fortiss.tooling.base.layout.util.OrientationUtils;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.kernel.base.ui.dnd.DropContext;

/**
 * Utility class collecting EditPart layout methods.
 * 
 * @author wandinger
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 3964EB0B7D1E388B567D8C31BCBD16F2
 */
public class EditPartLayoutUtils {

	/** Returns the given value aligned to the grid. */
	public static int snapToGrid(double d) {
		return RecommendedLayoutConstants.GRID_SIZE
				* (int) Math.round(d / RecommendedLayoutConstants.GRID_SIZE);
	}

	/** Modifies the x and y coordinates of the given point to be on the grid. */
	public static void snapToGrid(Point p) {
		p.x = snapToGrid(p.x);
		p.y = snapToGrid(p.y);
	}

	/**
	 * Returns polar coordinates of the point on the ellipse border having the
	 * given angle.
	 */
	public static Polar calculateEllipseBorderPolar(double angle,
			Dimension ellipseSize, int inset) {
		double a = ellipseSize.width / 2 - inset;
		double b = ellipseSize.height / 2 - inset;
		double epsilon = Math.sqrt(a * a - b * b) / a;
		double r = b
				/ Math.sqrt(1 - epsilon * epsilon * Math.cos(angle)
						* Math.cos(angle));
		return new Polar(angle, r);
	}

	/**
	 * Returns Cartesian coordinates of the point on the ellipse border having
	 * the given angle.
	 */
	public static Point calculateEllipseBorderPoint(double angle,
			Dimension ellipseSize, int inset) {
		Polar p = calculateEllipseBorderPolar(angle, ellipseSize, inset);
		return p.toCartesian().translate(ellipseSize.getScaled(.5));
	}

	/**
	 * Returns the {@link OffsetOrientation} of the point on the rectangle
	 * border having a given Point inside the rectangle.
	 */
	public static OffsetOrientation calculateRectangleBorderOrientation(
			Point source, Dimension rectangleSize, int inset) {
		int offset = 0;
		EOrientation orientation = EOrientation.NORTH;

		// The figure bounds
		Rectangle bounds = new Rectangle(0, 0, rectangleSize.width,
				rectangleSize.height);

		// Divide the rectangle into four diagonal sectors
		Polygon northSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.x, bounds.width }, new int[] { bounds.getCenter().y,
				bounds.y, bounds.y }, 3);
		Polygon eastSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.width, bounds.width }, new int[] { bounds.getCenter().y,
				bounds.y, bounds.height }, 3);
		Polygon southSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.width, bounds.x }, new int[] { bounds.getCenter().y,
				bounds.height, bounds.height }, 3);
		Polygon westSector = new Polygon(new int[] { bounds.getCenter().x,
				bounds.x, bounds.x }, new int[] { bounds.getCenter().y,
				bounds.height, bounds.y }, 3);

		// Determine in which sector we are and clamp the position accordingly
		if (northSector.contains(source.x, source.y)) {
			orientation = EOrientation.NORTH;
			offset = (int) (source.x * (1 - (2 * inset) / (double) bounds.width));
		}
		if (westSector.contains(source.x, source.y)) {
			orientation = EOrientation.WEST;
			offset = (int) (source.y * (1 - (2 * inset)
					/ (double) bounds.height));
		}
		if (southSector.contains(source.x, source.y)) {
			orientation = EOrientation.SOUTH;
			offset = (int) (source.x * (1 - (2 * inset) / (double) bounds.width));
		}
		if (eastSector.contains(source.x, source.y)) {
			orientation = EOrientation.EAST;
			offset = (int) (source.y * (1 - (2 * inset)
					/ (double) bounds.height));
		}
		return OrientationUtils.createOffsetOrientation(orientation, offset);
	}

	/**
	 * Ensure that the minimum size and aspect ratio of a oval shape are
	 * redeemed.
	 */
	public static Dimension clampOvalShapeSize(Dimension oldSize, int direction) {
		Dimension newSize;

		// check for minimum size
		newSize = new Dimension(
				oldSize.width < RecommendedLayoutConstants.OVAL_SHAPE_MINIMUM_WIDTH ? RecommendedLayoutConstants.OVAL_SHAPE_MINIMUM_WIDTH
						: oldSize.width,
				oldSize.height < RecommendedLayoutConstants.OVAL_SHAPE_MINIMUM_HEIGHT ? RecommendedLayoutConstants.OVAL_SHAPE_MINIMUM_HEIGHT
						: oldSize.height);

		// ensure that the aspect ratio is redeemed
		switch (direction) {
		case PositionConstants.SOUTH_EAST:
			int aspectHeight = (int) (newSize.width / RecommendedLayoutConstants.OVAL_SHAPE_ASPECT_RATIO);
			int aspectWidth = (int) (newSize.height * RecommendedLayoutConstants.OVAL_SHAPE_ASPECT_RATIO);
			newSize.setSize(new Dimension(
					newSize.width > aspectWidth ? aspectWidth : newSize.width,
					newSize.height > aspectHeight ? aspectHeight
							: newSize.height));
			break;
		case PositionConstants.SOUTH:
			newSize.setSize(new Dimension(
					(int) (newSize.height * RecommendedLayoutConstants.OVAL_SHAPE_ASPECT_RATIO),
					newSize.height));
			break;
		case PositionConstants.EAST:
			newSize.setSize(new Dimension(
					newSize.width,
					(int) (newSize.width / RecommendedLayoutConstants.OVAL_SHAPE_ASPECT_RATIO)));
			break;
		}

		return newSize;
	}

	/**
	 * Ensure that the minimum size of a rectangular shape is redeemed.
	 */
	public static Dimension clampRectangularShapeSize(Dimension oldSize) {
		// check for minimum size
		Dimension newSize = new Dimension(
				oldSize.width < RecommendedLayoutConstants.RECTANGULAR_SHAPE_MINIMUM_WIDTH ? RecommendedLayoutConstants.RECTANGULAR_SHAPE_MINIMUM_WIDTH
						: oldSize.width,
				oldSize.height < RecommendedLayoutConstants.RECTANGULAR_SHAPE_MINIMUM_HEIGHT ? RecommendedLayoutConstants.RECTANGULAR_SHAPE_MINIMUM_HEIGHT
						: oldSize.height);

		return newSize;
	}

	/** Layouts the given connector within the surrounding ellipse. */
	public static void layoutConnectorInEllipse(ILayoutedModelElement layouted,
			ILayoutedModelElement container, Point location) {
		Point position = location.getCopy();
		position.translate(LayoutDataUtils.getNodePosition(container)
				.getNegated());
		position.translate(LayoutDataUtils.getNodeSize(container).scale(-.5));
		// store the value in the model
		LayoutDataUtils.setAngle(layouted, LayoutKeyConstants.CONNECTOR_ANGLE,
				new Polar(position).angle);
	}

	/** Layouts the given connector within the surrounding rectangle. */
	public static void layoutConnectorInRectangle(
			ILayoutedModelElement layouted, ILayoutedModelElement container,
			Point location) {
		Point position = location.getCopy();
		position.translate(LayoutDataUtils.getNodePosition(container)
				.getNegated());
		// do border clamping
		OffsetOrientation orient = EditPartLayoutUtils
				.calculateRectangleBorderOrientation(position,
						LayoutDataUtils.getNodeSize(container),
						RecommendedLayoutConstants.RECTANGULAR_SHAPE_INSETS);
		// snap to grid and store the value in the model
		LayoutDataUtils.setConnectorOrientation(layouted,
				orient.getOrientation());
		LayoutDataUtils.setConnectorOffset(layouted,
				EditPartLayoutUtils.snapToGrid(orient.getOffset()));
	}

	/** Layouts the given layouted object using the position in the context. */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted,
			DropContext context, int positionOffset, boolean snapToGrid) {
		layoutNodeInRoot(layouted, context.getPosition(), positionOffset,
				snapToGrid);
	}

	/** Layouts the given layouted object using the position in the context. */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted,
			Point location, int positionOffset, boolean snapToGrid) {
		Point position = location.getCopy();
		position.translate(-positionOffset, -positionOffset);
		if (snapToGrid) {
			EditPartLayoutUtils.snapToGrid(position);
		}
		LayoutDataUtils.setNodePosition(layouted, position);
	}

	/**
	 * Class for storing polar coordinates.
	 * 
	 * @author wandinger
	 * @author hummel
	 * @author $Author$
	 * @version $Rev$
	 * @levd.rating YELLOW Rev: 519
	 */
	public static class Polar {

		/** The angle */
		public double angle;

		/** The distance */
		public double distance;

		/** The constructor */
		public Polar(double angle, double distance) {
			this.angle = angle;
			this.distance = distance;
		}

		/** Create new polar values from cartesian ones. */
		public Polar(Point pos) {
			angle = Math.atan2(pos.y, pos.x);
			distance = Math.sqrt(pos.x * pos.x + pos.y * pos.y);
		}

		/** Returns a cartesian representation. */
		public Point toCartesian() {
			return new Point(distance * Math.cos(angle), distance
					* Math.sin(angle));
		}
	}
}
